@@ -3,24 +3,31 @@
 
 use std::path::PathBuf;
 
+use colored::Colorize;
 use log::LevelFilter;
-use monument_cli::DebugOption;
+use monument::Config;
+use monument_cli::{CtrlCBehaviour, DebugOption};
 use structopt::StructOpt;
 
-/// Max number of comp prefixes stored in the queues of all threads
-const DEFAULT_QUEUE_LIMIT: usize = 10_000_000;
-
 fn main() {
     let args = CliArgs::from_args();
     monument_cli::init_logging(args.log_level());
-    let maybe_results = monument_cli::run(
-        &args.input_file,
-        args.debug,
-        args.queue_limit.unwrap_or(DEFAULT_QUEUE_LIMIT),
-    )
-    .unwrap();
-    if let Some(results) = maybe_results {
-        results.print();
+    let result = monument_cli::run(
+        monument_cli::Source::Path(&args.input_file),
+        args.debug_option,
+        &args.config(),
+        CtrlCBehaviour::RecoverComps,
+    );
+    match result {
+        Ok(Some(query_result)) => query_result.print(),
+        Ok(None) => assert!(args.debug_option.is_some()),
+        Err(e) => {
+            // In the case of an error, print the error message nicely then terminate the program
+            // with code -1 without causing a panic message.
+            println!("{} {:?}", "Error:".bright_red().bold(), e);
+            drop(args);
+            std::process::exit(-1);
+        }
     }
 }
 
@@ -28,35 +35,39 @@ fn main() {
 /// parsing/help code for us.
 #[derive(Debug, Clone, StructOpt)]
 #[structopt(name = "Monument", about = "Fast and flexible composing engine")]
-pub struct CliArgs {
+struct CliArgs {
     /// The name of the specification file for Monument (`*.toml`)
     #[structopt(parse(from_os_str))]
-    pub input_file: PathBuf,
+    input_file: PathBuf,
 
     /// The maximum number of threads that Monument will use.
     // TODO: Uncomment this once multi-threading is possible
     // #[structopt(short = "T", long)]
-    pub num_threads: Option<usize>,
+    num_threads: Option<usize>,
+    /// The maximum number of chunks in the chunk graph.  Exceeding this during generation will
+    /// cause an error.  Defaults to 100K.
+    #[structopt(long)]
+    graph_size_limit: Option<usize>,
     /// The maximum number of prefixes that Monument will store at once.  Defaults to 10 million.
     #[structopt(short = "Q", long)]
-    pub queue_limit: Option<usize>,
+    queue_limit: Option<usize>,
 
     /// Makes Monument print more output (`-vv` will produce all output).
     #[structopt(short, long = "verbose", parse(from_occurrences))]
-    pub verbosity: usize,
+    verbosity: usize,
     /// Makes Monument print less output (`-qq` will only produce errors).
     #[structopt(short, long = "quiet", parse(from_occurrences))]
-    pub quietness: usize,
+    quietness: usize,
 
     /// Debug options.  `spec`, `query`, `layout` and `graph` print the corresponding data
     /// structures.  `no-search` will run as normal but stop just before starting the full search.
     #[structopt(short = "D", long)]
-    pub debug: Option<DebugOption>,
+    debug_option: Option<DebugOption>,
 }
 
 impl CliArgs {
     /// Parse the `-q`/`-v` args into the [`LevelFilter`] to give to the `log` library
-    pub fn log_level(&self) -> LevelFilter {
+    fn log_level(&self) -> LevelFilter {
         match self.verbosity as isize - self.quietness as isize {
             x if x < -2 => LevelFilter::Off, // -qqq (or more `q`s)
             -2 => LevelFilter::Error,        // -qq
@@ -67,4 +78,22 @@ impl CliArgs {
             _ => LevelFilter::Trace,         // -vvv (or more `v`s)
         }
     }
+
+    fn config(&self) -> Config {
+        let mut config = Config {
+            num_threads: self.num_threads,
+            // Don't `drop` any of the search data structures, since Monument will exit shortly
+            // after the search terminates.  With the `Arc`-based data structures, this is
+            // seriously beneficial - it shaves many seconds off Monument's total running time.
+            mem_forget_search_data: true,
+            ..Config::default()
+        };
+        if let Some(q) = self.queue_limit {
+            config.queue_limit = q;
+        }
+        if let Some(g) = self.graph_size_limit {
+            config.graph_size_limit = g;
+        }
+        config
+    }
 }
